_G["LootSquid"] = {
    components = {},
    session = {
        loot = {},
        currentLootSlot = 0,
        playersNeeding = {},
        playersGreeding = {}
    },
    debug = { alwaysParseLoot = false, alwaysParseChat = false, logging = false, disableWhispers = false },
    settings = { lootThreshold = 0 },
    frames = { mainFrame = nil },
    eventHandlers = {},
    data = { maxStanding = 0 },
    lastChatRequestEventTime = 0,
    lastGetItemInfoEventTime = 0,
    lastRequestFrameUpdatedTime = 0,
    versionMajor = "0",
    versionMinor = "3",
    versionRevision = "5",
    versionState = "beta"
}

StaticPopupDialogs["LOOTSQUID_AWARD_LOOT"] = {
    text = "Transfer loot?",
    button1 = "Yes",
    button2 = "No",
    OnAccept = function(popup)

    -- find candidate for current loot slot
        local shortPlayerName = Ambiguate(popup.playerName, "none")
        for k in pairs(_G.LootSquid.session.loot) do
            if (_G.LootSquid.session.loot[k].candidates
                    and _G.LootSquid.session.loot[k].candidates[shortPlayerName]) then

                local ci = _G.LootSquid.session.loot[k].candidates[shortPlayerName].candidateIndex

                if ci > 0 then
                    LootSquid:TransferLoot(_G.LootSquid.session.currentLootSlot, ci)
                    LootSquid:ClearNeedAndGreedRequests()
                end
            end
        end

    --
    end,
    timeout = 30,
    whileDead = false,
    hideOnEscape = true,
    preferredIndex = 3, -- avoid some UI taint, see http://www.wowace.com/announcements/how-to-avoid-some-ui-taint/
}

function LootSquid:OnLoad(mainFrame)
    mainFrame:ClearAllPoints()
    mainFrame:SetPoint("LEFT", 0)
    mainFrame:SetPoint("TOP", 0)
    self.frames.mainFrame = mainFrame
    self:Initialize()
end

function LootSquid:Initialize()

    local components = {
        ["LootSquid Data"] = LootSquidData,
        ["LootSquid Utilities"] = LootSquid.components.utils,
        ["LootSquid Events"] = LootSquid.components.events,
        ["LootSquid Information Logging"] = LootSquid.LogInfo,
        ["LootSquid Error Logging"] = LootSquid.LogError
    }

    for k,v in pairs(components) do
        if not v then
            DEFAULT_CHAT_FRAME:AddMessage(
                format(
                    "|cFFFFAA00%s:|cFFFF0000 %s", "LootSquid [Error]",
                    "Missing Component: "
                            ..tostring(k)
                            ..". Please make sure addon is installed correctly."
                ))
        end
    end


    self:RegisterEvents()

    --    precalculate api rating
    if _G["LootSquidData"] then
        if (_G["LootSquidData"].standings) then
            for k in pairs(_G["LootSquidData"].standings) do
                local standing = _G["LootSquidData"].standings[k].standing
                local itemsReceived = _G["LootSquidData"].standings[k].itemsReceived
                local apiRating = _G["LootSquidData"].standings[k].apiRating

                self.data.maxStanding = math.max(standing, self.data.maxStanding)

                if (standing and itemsReceived) then
                    if (itemsReceived > 0) then
                        apiRating = standing / itemsReceived
                    else
                        apiRating = standing / 0.5
                    end
                    _G["LootSquidData"].standings[k].apiRating = apiRating
                end
            end
        else
            LootSquid:LogError("No standings available in LootSquidData.")
        end
    else
        LootSquid:LogError("No standings available. Probably missing LootSquidData.lua file.")
    end
end

function LootSquid:TimerFeedback()
    print("5 seconds passed")
end

function LootSquid:LootIconClick(lootIndex)
    --LootSquid:LogInfo("Selecting lootIndex: " .. tostring(lootIndex))
end

function LootSquid:LootIconEnter(itemLink)
    GameTooltip:SetOwner(self.frames.window, "ANCHOR_CURSOR")
    GameTooltip:SetHyperlink(itemLink)
    GameTooltip:Show()
end

function LootSquid:LootIconLeave()
    GameTooltip:Hide()
end

function LootSquid:TransferLoot(lootSlot, candidateIndex)
    GiveMasterLoot(lootSlot, candidateIndex)
end

function LootSquid:CreateBackdropSolid(edgeSize, insets)

    local edgeFile = "Interface\\ChatFrame\\ChatFrameBackground"

    -- backdrops can't handle edgesize of zero with edgeFile defined
    if edgeSize < 1 then
        edgeFile = nil
    end

    local backdrop = {
        bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
        edgeFile = edgeFile,
        tile = true,
        -- size (width or height) of the square repeating background tiles (in pixels)
        tileSize = 32,
        -- thickness of edge segments and square size of edge corners (in pixels)
        edgeSize = edgeSize,
        -- distance from the edges of the frame to those of the background texture (in pixels)
        insets = insets
    }

    return backdrop
end

function LootSquid:CreateButtonStandard(parent, width, height, text)
    local b = CreateFrame("Button", nil, parent, "UIPanelButtonTemplate")
    b:SetText(text)
    b:SetWidth(width)
    b:SetHeight(height)
    b:ClearAllPoints()
    return b
end

function LootSquid:CreateFrameSolid(parent, width, height, backgroundColor, edgeColor, edgeSize, isMoveAble)
    local f = CreateFrame("Frame", nil, parent)
    if width then
        f:SetWidth(width)
    end
    if height then
        f:SetHeight(height)
    end
    f:SetBackdrop(self:CreateBackdropSolid(edgeSize))
    f:SetBackdropColor(backgroundColor.r, backgroundColor.g, backgroundColor.b, backgroundColor.a)
    f:SetBackdropBorderColor(edgeColor.r, edgeColor.g, edgeColor.b, edgeColor.a)

    if isMoveAble then

        f:SetMovable(true)
        f:EnableMouse(true)

        f:SetScript("OnMouseDown",
            function(selfFrame, button)
                if button == "LeftButton" and not selfFrame.isMoving then
                    selfFrame:StartMoving()
                    selfFrame.isMoving = true
                end
            end)

        f:SetScript("OnMouseUp",
            function(selfFrame, button)
                if button == "LeftButton" and selfFrame.isMoving then
                    selfFrame:StopMovingOrSizing()
                    selfFrame.isMoving = false
                end
            end)

        f:SetScript("OnHide",
            function(selfFrame, button)
                if selfFrame.isMoving then
                    selfFrame:StopMovingOrSizing()
                    selfFrame.isMoving = false
                end
            end)
    end

    return f
end

function LootSquid:CreateLabelStandard(parent, text, textHeight, font, isWrapDisabled)

    if font == nil then
        font = "GameFontNormal"
    end

    local label = parent:CreateFontString(nil, "OVERLAY", font)
    label:SetText(text)
    if textHeight ~= nil then
        label:SetTextHeight(textHeight)
    end

    if isWrapDisabled == true then
        label:SetWordWrap(false)
    else
        label:SetWordWrap(true)
    end

    return label
end

function LootSquid:CreateScrollFrame(frame, itemHeight, itemCnt, padding, funcItem)

    local result = {}

    local frameScroll = CreateFrame("ScrollFrame", nil, frame)
    frameScroll:SetPoint("TOPLEFT", 10, -20)
    frameScroll:SetPoint("BOTTOMRIGHT", -10, 10)

    result.frameScroll = frameScroll

    local texture = frameScroll:CreateTexture()
    texture:SetAllPoints()
    texture:SetTexture(0, 0, 0, 1)

    local scrollbar = CreateFrame("Slider", nil, frameScroll, "UIPanelScrollBarTemplate")
    scrollbar:SetPoint("TOPLEFT", frame, "TOPRIGHT", 4, -16)
    scrollbar:SetPoint("BOTTOMLEFT", frame, "BOTTOMRIGHT", 4, 16)

    --local scrollBarMax = math.max((itemHeight+padding) * itemCnt - frameScroll:GetHeight() - padding, 1)
    local scrollBarMax = math.max((itemHeight + padding) * itemCnt - padding, 1)

    scrollbar:SetMinMaxValues(0, scrollBarMax + 1)
    scrollbar:SetValueStep(itemHeight + padding)
    scrollbar.scrollStep = itemHeight + padding
    scrollbar:SetValue(0)
    scrollbar:SetWidth(16)
    scrollbar:SetScript("OnValueChanged",
        function(self, value)
            self:GetParent():SetVerticalScroll(value)
        end)
    scrollbar:EnableMouseWheel(true)
    result.scrollbar = scrollbar

    frameScroll:SetScript("OnMouseWheel",
        function(self, delta)

            local current = self:GetVerticalScroll()
            local min, max = scrollbar:GetMinMaxValues()

            local offsetCnt = math.floor(current / (itemHeight + padding))
            offsetCnt = offsetCnt - delta

            local offset = offsetCnt * (itemHeight + padding)

            if offset >= min and offset <= max then
                scrollbar:SetValue(offset)
            end
        end)

    local scrollbg = scrollbar:CreateTexture(nil, "BACKGROUND")
    scrollbg:SetAllPoints(scrollbar)
    scrollbg:SetTexture(0, 0, 0, 0.4)

    local lootFrameScrollChild = CreateFrame("Frame", nil, frameScroll)
    lootFrameScrollChild:SetWidth(frameScroll:GetWidth())
    lootFrameScrollChild:SetHeight(itemHeight * itemCnt)

    frameScroll:SetScrollChild(lootFrameScrollChild)


    for i = 0, itemCnt - 1 do
        local lootItemHolder = self:CreateFrameSolid(lootFrameScrollChild,
            lootFrameScrollChild:GetWidth(),
            itemHeight,
            { r = 0.1, g = 0.1, b = 0.1, a = 1.0 },
            { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
            0,
            false)
        lootItemHolder:SetPoint("TOP", 0, -(padding + itemHeight) * i)

        funcItem(lootItemHolder, i)
    end


    result.HideIndexAndAfter = function(self, frameScroll, index)

        for i = index, LootSquid:tablelength(frameScroll.items) - 1 do
            frameScroll.items[i].frame:Hide()
        end

        local scrollBarMax = math.max((itemHeight + padding) * index - padding, 1)
        frameScroll.scrollbar:SetMinMaxValues(0, scrollBarMax + 1)
    end


    local frameMessage = self:CreateLabelStandard(frameScroll, "No more items", nil, "GameFontWhiteTiny")
    frameMessage:SetPoint("TOP", 0, -20)

    result.frameMessage = frameMessage

    return result
end

function LootSquid:InitGui()
    -- create new window

    -- settings
    local headerHeight = 31
    local footerHeight = 20
    local padding = 3

    local f = self:CreateFrameSolid(UIParent,
        --GetScreenWidth() * 0.6,
        --GetScreenHeight() * 0.6,
        1200,
        800,
        { r = 0.1, g = 0.1, b = 0.1, a = 1.0 },
        { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
        padding,
        true)

    f:SetPoint("CENTER", 0, 0)
    f:SetFrameStrata("FULLSCREEN")

    f:SetScript("OnKeyDown",
        function(self, key)
            if key == "ESCAPE" then
                f:SetPropagateKeyboardInput(false)
            else
                f:SetPropagateKeyboardInput(true)
            end
        end
    )

    f:SetScript("OnKeyUp",
        function(self, key)
            if key == "ESCAPE" then
                f:Hide()
            end
        end
        )

    local header = self:CreateFrameSolid(f,
        f:GetWidth() - padding * 2,
        headerHeight,
        { r = 0.12, g = 0.12, b = 0.12, a = 1.0 },
        { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
        0,
        false)
    header:SetPoint("TOP", 0, -padding)
    header:SetPoint("LEFT", padding, 0)

    local b = self:CreateButtonStandard(header, 25, 25, "x")
    b:SetPoint("RIGHT", -padding, 0)
    b:SetPoint("TOP", 0, -padding)
    b:SetScript("OnClick", function(selfButton, button) f:Hide() end)

    local title = self:CreateLabelStandard(header,
        string.format("Loot Squid - v%d.%d.%d %s", self.versionMajor, self.versionMinor, self.versionRevision, self.versionState))
    title:SetPoint("LEFT", 8, 0)

    local footer = self:CreateFrameSolid(f,
        f:GetWidth() - 6,
        footerHeight,
        { r = 0.12, g = 0.12, b = 0.12, a = 1.0 },
        { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
        0,
        false)
    footer:SetPoint("BOTTOM", 0, padding)
    footer:SetPoint("LEFT", padding, 0)

    local bodyFrame = self:CreateFrameSolid(f,
        (f:GetWidth() - padding * 2),
        (f:GetHeight() - padding * 4 - headerHeight - footerHeight),
        { r = 0.1, g = 0.1, b = 0.1, a = 1.0 },
        { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
        0,
        false)
    bodyFrame:SetPoint("TOP", 0, -padding - headerHeight - padding)
    bodyFrame:SetPoint("LEFT", padding, 0)

    local lootFrame = self:CreateFrameSolid(bodyFrame,
        (bodyFrame:GetWidth() - padding) / 2 - 16 - padding,
        (bodyFrame:GetHeight() - padding * 2) / 3,
        { r = 0.2, g = 0.2, b = 0.2, a = 1.0 },
        { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
        0,
        false)
    lootFrame:SetPoint("TOP", 0, 0)
    lootFrame:SetPoint("LEFT", 0, 0)
    local lootFrameTitle = self:CreateLabelStandard(lootFrame,
        "Loot - " .. tostring(GetUnitName("target", false)),
        nil, "GameFontWhiteTiny")
    lootFrameTitle:SetPoint("LEFT", 10, 0)
    lootFrameTitle:SetPoint("TOP", 0, -7)
    self.frames.lootFrameTitle = lootFrameTitle

    local lootFrameScrollItems = {}
    local lootFrameScroll = self:CreateScrollFrame(lootFrame, 100, 20, 3,
        function(frame, index)

            local textureAndLabelWrapper = CreateFrame("Frame", nil, frame)
            textureAndLabelWrapper:SetSize(80, 80)
            textureAndLabelWrapper:SetPoint("LEFT", 25, 0)

            local texture = textureAndLabelWrapper:CreateTexture()
            texture:SetSize(40, 40)
            texture:SetPoint("CENTER", 0, 10)
            texture:SetTexture(1, 1, 1, 1)

            local label = self:CreateLabelStandard(textureAndLabelWrapper,
                "Not initialized",
                nil, "GameFontWhiteTiny")
            label:SetPoint("CENTER", 0, -25)
            label:SetWidth(80)

            -- parent, width, height, text
            local offerButton = LootSquid:CreateButtonStandard(frame, 150, 30, "Offer to Raid")
            offerButton:SetPoint("RIGHT", -20, 0)
            offerButton:SetPoint("TOP", 0, -20)

            local transferButton = LootSquid:CreateButtonStandard(frame, 150, 30, "Transfer to me")
            transferButton:SetPoint("RIGHT", -20, 0)
            transferButton:SetPoint("BOTTOM", 0, 20)

            lootFrameScrollItems[index] = {
                frame = frame,
                texture = texture,
                label = label,
                offerButton = offerButton,
                transferButton = transferButton,
                textureAndLabelWrapper = textureAndLabelWrapper
            }
        end);

    lootFrameScroll.items = lootFrameScrollItems

    self.frames.lootFrameScroll = lootFrameScroll;


    local overviewFrame = self:CreateFrameSolid(bodyFrame,
        (bodyFrame:GetWidth() - padding) / 2,
        (bodyFrame:GetHeight() - padding * 2) / 3,
        { r = 0.2, g = 0.2, b = 0.2, a = 1.0 },
        { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
        0,
        false)
    overviewFrame:SetPoint("TOP", 0, 0)
    overviewFrame:SetPoint("RIGHT", 0, 0)

    local overviewLabel = self:CreateLabelStandard(overviewFrame,
        "Overview",
        nil, "GameFontWhiteTiny")
    overviewLabel:SetPoint("LEFT", overviewFrame, "TOPLEFT", 10, -10)


    local overviewCurrentSessionItemName = self:CreateLabelStandard(overviewFrame,
        "No session is active",
        nil, "GameFontWhiteTiny")
    overviewCurrentSessionItemName:SetPoint("CENTER", overviewFrame, "TOPLEFT", 120, -40)

    self.frames.overviewCurrentSessionItemName = overviewCurrentSessionItemName


    local playerNeedButton = LootSquid:CreateButtonStandard(overviewFrame, 150, 30, "Need it")
    playerNeedButton:SetPoint("CENTER", overviewFrame, "TOPLEFT", 120, -110)
    playerNeedButton:SetScript("OnClick",
        function()
            if (_G.LootSquid.session.currentLootSlot < 1) then
                LootSquid:LogError("No session is active. Cannot need an item.")
            else
                _G.LootSquid.session.playersGreeding[UnitName("player")] = nil
                _G.LootSquid.session.playersNeeding[UnitName("player")] =
                { isNeeding = true, currentItemLink = nil }
                LootSquid:UpdateRequestFrame()
            end
        end)

    local playerGreedButton = LootSquid:CreateButtonStandard(overviewFrame, 150, 30, "Greed it")
    playerGreedButton:SetPoint("CENTER", overviewFrame, "TOPLEFT", 120, -150)
    playerGreedButton:SetScript("OnClick",
        function()
            if (_G.LootSquid.session.currentLootSlot < 1) then
                LootSquid:LogError("No session is active. Cannot greed an item.")
            else
                _G.LootSquid.session.playersNeeding[UnitName("player")] = nil
                _G.LootSquid.session.playersGreeding[UnitName("player")] =
                { isGreeding = true, currentItemLink = nil }
                LootSquid:UpdateRequestFrame()
            end
        end)




    local requestFrame = self:CreateFrameSolid(bodyFrame,
        nil, --(bodyFrame:GetWidth()),
        nil, --(bodyFrame:GetHeight() - padding * 2) * 2 / 3 + padding,
        { r = 0.2, g = 0.2, b = 0.2, a = 1.0 },
        { r = 0.0, g = 0.0, b = 0.0, a = 1.0 },
        0,
        false)
    requestFrame:SetPoint("TOPLEFT", 0, -(bodyFrame:GetHeight() - padding * 2) / 3 - padding)
    requestFrame:SetPoint("BOTTOMRIGHT", -16 - padding, 0)



    local requestHeaderLabel = self:CreateLabelStandard(requestFrame,
        "Player info",
        nil, "GameFontWhiteTiny")
    requestHeaderLabel:SetPoint("CENTER", requestFrame, "TOPLEFT", 100, -10)

    requestHeaderLabel = self:CreateLabelStandard(requestFrame,
        "Current Item",
        nil, "GameFontWhiteTiny")
    requestHeaderLabel:SetPoint("CENTER", requestFrame, "TOPLEFT", 200, -10)

    requestHeaderLabel = self:CreateLabelStandard(requestFrame,
        "Improvement",
        nil, "GameFontWhiteTiny")
    requestHeaderLabel:SetPoint("CENTER", requestFrame, "TOPLEFT", 350, -10)

    requestHeaderLabel = self:CreateLabelStandard(requestFrame,
        "Attendance",
        nil, "GameFontWhiteTiny")
    requestHeaderLabel:SetPoint("CENTER", requestFrame, "TOPLEFT", 570, -10)

    requestHeaderLabel = self:CreateLabelStandard(requestFrame,
        "Items Received",
        nil, "GameFontWhiteTiny")
    requestHeaderLabel:SetPoint("CENTER", requestFrame, "TOPLEFT", 720, -10)

    requestHeaderLabel = self:CreateLabelStandard(requestFrame,
        "API Rating",
        nil, "GameFontWhiteTiny")
    requestHeaderLabel:SetPoint("CENTER", requestFrame, "TOPLEFT", 870, -10)

    requestHeaderLabel = self:CreateLabelStandard(requestFrame,
        "Actions",
        nil, "GameFontWhiteTiny")
    requestHeaderLabel:SetPoint("CENTER", requestFrame, "TOPLEFT", 1050, -10)



    local requestFrameScrollItems = {}
    local requestFrameScroll = self:CreateScrollFrame(requestFrame, 100, 20, 3,
        function(frame, index)

            local labelName = self:CreateLabelStandard(frame,
                "labelName",
                nil, "GameFontWhiteTiny")
            labelName:SetPoint("CENTER", frame, "TOPLEFT", 80, -20)

            local labelSpec = self:CreateLabelStandard(frame,
                "labelSpec",
                nil, "GameFontWhiteTiny")
            labelSpec:SetPoint("CENTER", frame, "LEFT", 80, 0)

            local labelStatPrio = self:CreateLabelStandard(frame,
                "labelStatPrio",
                nil, "GameFontWhiteTiny")
            labelStatPrio:SetPoint("CENTER", frame, "BOTTOMLEFT", 80, 20)

            local textureAndLabelWrapper = CreateFrame("Frame", nil, frame)
            textureAndLabelWrapper:SetSize(80, 80)
            textureAndLabelWrapper:SetPoint("LEFT", 150, 0)

            local textureItem = textureAndLabelWrapper:CreateTexture()
            textureItem:SetSize(40, 40)
            textureItem:SetPoint("CENTER", 0, 10)
            textureItem:SetTexture(1, 1, 1, 1)

            local labelItem = self:CreateLabelStandard(textureAndLabelWrapper,
                "Not initialized",
                nil, "GameFontWhiteTiny")
            labelItem:SetPoint("CENTER", 0, -25)
            labelItem:SetWidth(80)

            local labelItemLvlIncrease = self:CreateLabelStandard(frame,
                "labelItemLvlIncrease",
                nil, "GameFontWhiteTiny")
            labelItemLvlIncrease:SetPoint("CENTER", frame, "LEFT", 350, 0)
            labelItemLvlIncrease:SetWidth(100)

            local labelAttendance12 = self:CreateLabelStandard(frame,
                "labelAttendance12",
                nil, "GameFontWhiteTiny")
            labelAttendance12:SetPoint("CENTER", frame, "RIGHT", -550, 0)

            local labelItems12 = self:CreateLabelStandard(frame,
                "labelItems12",
                nil, "GameFontWhiteTiny")
            labelItems12:SetPoint("CENTER", frame, "RIGHT", -400, 0)

            local labelApi = self:CreateLabelStandard(frame,
                "labelApi",
                nil, "GameFontWhiteTiny")
            labelApi:SetPoint("CENTER", frame, "RIGHT", -250, 0)


            local labelNeedGreed = self:CreateLabelStandard(frame,
                "NeedOrGreed?",
                nil, "GameFontWhiteTiny")
            labelNeedGreed:SetPoint("CENTER", frame, "TOPRIGHT", -80, -15)

            -- parent, width, height, text
            local awardButton = LootSquid:CreateButtonStandard(frame, 150, 30, "Award")
            awardButton:SetPoint("CENTER", frame, "RIGHT", -80, 8)

            local rollButton = LootSquid:CreateButtonStandard(frame, 150, 30, "Whisper for roll")
            rollButton:SetPoint("CENTER", frame, "BOTTOMRIGHT", -80, 25)

            requestFrameScrollItems[index] = {
                frame = frame,
                labelName = labelName,
                labelSpec = labelSpec,
                labelStatPrio = labelStatPrio,
                textureAndLabelWrapper = textureAndLabelWrapper,
                textureItem = textureItem,
                labelItem = labelItem,
                labelItems12 = labelItems12,
                labelAttendance12 = labelAttendance12,
                labelApi = labelApi,
                labelNeedGreed = labelNeedGreed,
                awardButton = awardButton,
                rollButton = rollButton,
                labelItemLvlIncrease = labelItemLvlIncrease
            }
        end);

    requestFrameScroll.items = requestFrameScrollItems

    self.frames.requestFrameScroll = requestFrameScroll;




    self.frames.window = f
end

function LootSquid:OpenLootGui()

    if self.frames.window == nil then
        self:InitGui()
    end

    -- attempt to detect if ElvUI has loot frame override enabled
    --    for i = 0, 10 do
    --        if ElvUI
    --                and ElvUI[i]
    --                and ElvUI[i].Skins
    --                and ElvUI[i].Skins.db
    --                and ElvUI[i].Skins.db.blizzard
    --                and ElvUI[i].Skins.db.blizzard.loot == true
    --        then
    --            LootSquid:LogInfo("|cFFFF0000 Warning: ElvUI found and has loot frame override enabled. "
    --            .. "Please go to ElvUI config > General and untick 'loot'. "
    --            .. "Optionally also concider disabling loot frame skinning in Skins > Loot Frames. "
    --            .. "The issue stems from ElvUI placing the loot frame in very high frame strata (above everything else).")
    --        end
    --    end


    local f = self.frames.window
    f:Show()

    self:UpdateLootFrame()
    self:UpdateRequestFrame()
end

function LootSquid:OnUpdateLootTable()
    self:UpdateLootFrame()
end

function LootSquid:UpdateCurrentOfferedLoot()

    --[[
    LootSquid.session.currentLootOffered = {
                            name = name,
                            link = link,
                            quality = quality,
                            iLevel = iLevel,
                            class = class,
                            subclass = subclass,
                            maxStack = maxStack,
                            equipSlot = equipSlot,
                            texture = texture
                        }
    -- ]]

    self.frames.overviewCurrentSessionItemName:SetText(LootSquid.session.currentLootOffered.name)
end

function LootSquid:UpdateLootFrame()
    if (not self.frames.lootFrameScroll) then
        return
    end

    local lootFrameScrollItemsIndex = 0

    for k in pairs(self.session.loot) do

        local v = self.session.loot[k]

        --[[
        -- loot[lootSlot] = {
            itemName = itemName,
            itemQuantity = itemQuantity,
            itemQuality = itemQuality,
            itemLocked = itemLocked,
            itemLootType = itemLootType,
            itemLink = itemLink,
            candidates = {}
        }]]

        --[[
         lootFrameScrollItems[index] = {
            frame = frame,
            texture = texture,
            label = label,
            offerButton = offerButton,
            transferButton = transferButton
        }]]

        local lootThreshold = GetLootThreshold()

        if v and v.itemLink and v.itemQuality >= lootThreshold then

            local name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice =
            GetItemInfo(v.itemLink)

            if self.frames.lootFrameScroll.items and
                    self.frames.lootFrameScroll.items[lootFrameScrollItemsIndex] then

                local lootFrameScrollItem = self.frames.lootFrameScroll.items[lootFrameScrollItemsIndex];

                lootFrameScrollItem.frame:Show()

                lootFrameScrollItem.textureAndLabelWrapper:EnableMouse(true)
                lootFrameScrollItem.textureAndLabelWrapper:SetScript("OnEnter", function(self, motion)
                    LootSquid:LootIconEnter(v.itemLink)
                end)
                lootFrameScrollItem.textureAndLabelWrapper:SetScript("OnLeave", function(self, motion)
                    LootSquid:LootIconLeave()
                end)

                lootFrameScrollItem.texture:SetTexture(texture)

                local r, g, b, hex = GetItemQualityColor(quality)

                lootFrameScrollItem.label:SetFormattedText("|c" .. hex .. name)

                lootFrameScrollItem.offerButton:RegisterForClicks("LeftButtonUp")
                lootFrameScrollItem.offerButton:SetScript("OnClick",
                    function(widget)
                        LootSquid:ClearNeedAndGreedRequests()
                        LootSquid.session.currentLootSlot = k
                        LootSquid.session.currentLootOffered = {
                            name = name,
                            link = link,
                            quality = quality,
                            iLevel = iLevel,
                            class = class,
                            subclass = subclass,
                            maxStack = maxStack,
                            equipSlot = equipSlot,
                            texture = texture
                        }
                        self:UpdateCurrentOfferedLoot()
                        LootSquid:RaidWarning("Loot session starting for: ".. v.itemLink)
                        SendChatMessage("LootSquid: do you want to need or greed "
                                .. v.itemLink .. " ? - if so, write either !need <currentItem> or !greed <currentItem> in chat. If you change your mind: write !pass",
                            "RAID")
                    end)

                lootFrameScrollItem.transferButton:RegisterForClicks("LeftButtonUp")
                lootFrameScrollItem.transferButton:SetScript("OnClick",
                    function(widget)
                        LootSquid.session.currentLootSlot = k
                        StaticPopupDialogs["LOOTSQUID_AWARD_LOOT"].text = "Really transfer item to you?"
                        local popup = StaticPopup_Show("LOOTSQUID_AWARD_LOOT", UnitName("player"))
                        if popup then
                            popup:SetFrameStrata("FULLSCREEN_DIALOG")
                            popup.playerName = UnitName("player")
                        end
                    end)
            end

            lootFrameScrollItemsIndex = lootFrameScrollItemsIndex + 1
        end
    end


    if lootFrameScrollItemsIndex < 1 then
        self.frames.lootFrameScroll.frameMessage:SetText("No more items available.")
        self.frames.lootFrameScroll.frameMessage:Show()
    else
        self.frames.lootFrameScroll.frameMessage:Hide()
    end

    LootSquid.frames.lootFrameScroll:HideIndexAndAfter(self.frames.lootFrameScroll, lootFrameScrollItemsIndex)
end

function LootSquid:GetStandingsByPlayerName(playerName)

--    local playerRealmName = GetRealmName()
--    local realmIndex = string.find(playerName, "-"..tostring(playerRealmName))
--
--    local playerNameFixed = playerName
--    if realmIndex then
--        playerNameFixed = string.sub(playerName, 1, realmIndex-1)
--        LootSquid:LogDebug("Fixing player name due to local realm to: "..tostring(playerName))
--    end

    local playerNameFixed = playerName
    local dashIndex = string.find(playerName, "-")
    if dashIndex then
        -- http://wow.gamepedia.com/API_Ambiguate
         playerNameFixed = Ambiguate(playerName, "none")
         LootSquid:LogDebug("Fixing player name due to local realm from: "..tostring(playerName).." to :"..tostring(playerNameFixed))
    end

    LootSquid:LogDebug("Standings lookup for player: "..tostring(playerNameFixed)..", dashIndex: "..tostring(dashIndex))

    if (_G["LootSquidData"] and _G["LootSquidData"].standings and _G["LootSquidData"].standings[playerNameFixed]) then
        local standing = _G["LootSquidData"].standings[playerNameFixed].standing
        local itemsReceived = _G["LootSquidData"].standings[playerNameFixed].itemsReceived
        local apiRating = _G["LootSquidData"].standings[playerNameFixed].apiRating
        return standing, itemsReceived, apiRating, playerNameFixed
    end
    return nil
end

function LootSquid:SortStandingsByApiRating(a, b)
    local aStanding, aItemsReceived, aApiRating = LootSquid:GetStandingsByPlayerName(a)
    local bStanding, bItemsReceived, bApiRating = LootSquid:GetStandingsByPlayerName(b)

    if aApiRating == nil and bApiRating == nil then
        return false
    end
    if aApiRating == nil then
        return false
    end
    if bApiRating == nil then
        return true
    end

    return aApiRating > bApiRating
end

function LootSquid:AddStandingsDataRows(requestFrameScrollItemsIndex, needingOrGreedingPlayers, isNeed)

    if (not self.frames.requestFrameScroll) then
        return -1
    end

    for k, v in self:spairs(needingOrGreedingPlayers,
        function(t, a, b) return self:SortStandingsByApiRating(a, b) end) do

        local standingPercent = 0

        local standing, itemsReceived, apiRating, playerNameFixed = LootSquid:GetStandingsByPlayerName(k)

        if (standing and itemsReceived and apiRating) then

            if LootSquid.data.maxStanding > 0 then
                standingPercent = (standing / LootSquid.data.maxStanding) * 100.0
            end
        else
            LootSquid:LogDebug("No standings for player name: " .. tostring(playerNameFixed))
            standing = 0
            itemsReceived = 0
            apiRating = 0
        end

        --[[
        --
        -- requestFrameScrollItems[index] = {
                frame = frame,
                labelName = labelName,
                labelSpec = labelSpec,
                labelStatPrio = labelStatPrio,
                textureItem = textureItem,
                labelItem = labelItem,
                labelItems12 = labelItems12,
                labelAttendance12 = labelAttendance12,
                labelApi = labelApi,
                labelNeedGreed = labelNeedGreed,
                awardButton = awardButton,
                rollButton = rollButton
            }
        --
        -- ]]


        if self.frames.requestFrameScroll.items and
                self.frames.requestFrameScroll.items[requestFrameScrollItemsIndex] then

            local requestFrameScrollItem = self.frames.requestFrameScroll.items[requestFrameScrollItemsIndex];

            local ambiguatedName = Ambiguate(k, "none")
            local class, classFileName = UnitClass(ambiguatedName)
            local color = RAID_CLASS_COLORS[classFileName]
            local playerNameColored
            if color then
                playerNameColored = string.format("|cFF%02X%02X%02X %s", color.r * 255, color.g * 255, color.b * 255, ambiguatedName)
            else
                playerNameColored = string.format("|cFF%02X%02X%02X %s", 0.5 * 255, 0.5 * 255, 0.5 * 255, ambiguatedName)
            end

            requestFrameScrollItem.frame:Show()
            requestFrameScrollItem.labelName:SetFormattedText(playerNameColored)


            requestFrameScrollItem.labelSpec:SetText(class)



            local standingColor = "FFFFFF"

            if LootSquid.data.maxStanding > 0 then
                if standing >= LootSquid.data.maxStanding * 0.8 then
                    standingColor = "00FF00" -- green
                elseif standing >= LootSquid.data.maxStanding * 0.6 then
                    standingColor = "9999FF" -- blue
                end
            end

            requestFrameScrollItem.labelAttendance12:SetFormattedText(
                string.format("|cFF%s %s (%.2f %s)", standingColor, tostring(standing), standingPercent, "%%"))
            requestFrameScrollItem.labelItems12:SetText(tostring(itemsReceived))
            requestFrameScrollItem.labelApi:SetText(string.format("%.2f", apiRating))

            if v.currentItemId then
                -- request the item info, if we already have it cached.
                -- An immidiate call after will either be nil (and pending server request) or contain the info.
                GetItemInfo(v.currentItemId)
            end

            if v.currentItemId then
                local currentItemName, currentItemLink, currentItemQuality,
                currentItemLevel, currentItemMinLevel, currentItemClass, currentItemSubclass,
                currentItemMaxStack, currentItemEquipSlot, currentItemTexture, currentItemVendorPrice =
                GetItemInfo(v.currentItemId)

                if (currentItemName and currentItemLink
                        and currentItemQuality and currentItemTexture
                        and currentItemLevel and currentItemEquipSlot) then

                    requestFrameScrollItem.textureAndLabelWrapper:EnableMouse(true)
                    requestFrameScrollItem.textureAndLabelWrapper:SetScript("OnEnter",
                        function(self, motion)
                        --[[ {
                            isNeeding = true,
                            currentItemId = Id
                            ]]

                            if (currentItemLink) then
                                LootSquid:LootIconEnter(currentItemLink)
                            end
                        end)

                    requestFrameScrollItem.textureAndLabelWrapper:SetScript("OnLeave",
                        function(self, motion)
                            LootSquid:LootIconLeave()
                        end)

                    local r, g, b, hex = GetItemQualityColor(currentItemQuality)
                    requestFrameScrollItem.labelItem:SetFormattedText("|c" .. hex .. currentItemName)


                    --LootSquid:LogDebug("Setting current item texture: " .. currentItemTexture)
                    requestFrameScrollItem.textureItem:SetTexture(currentItemTexture)


                    --[[
                        LootSquid.session.currentLootOffered = {
                            name = name,
                            link = link,
                            quality = quality,
                            lLevel = iLevel,
                            class = class,
                            subclass = subclass,
                            maxStack = maxStack,
                            equipSlot = equipSlot,
                            texture = texture
                        }
                    -- ]]

                    if LootSquid.session.currentLootOffered then

                        local iLevelDiff = LootSquid.session.currentLootOffered.iLevel - currentItemLevel

                        if LootSquid.session.currentLootOffered.equipSlot ~= currentItemEquipSlot then
                            requestFrameScrollItem.labelItemLvlIncrease:SetFormattedText("|cFFFF0000SLOT MISMATCH! "
                                    .. tostring(LootSquid.session.currentLootOffered.equipSlot)
                                    .. " vs " .. tostring(currentItemEquipSlot) .. ", diff: " .. tostring(iLevelDiff))
                        else

                            local prefixColor
                            if iLevelDiff < 0 then
                                prefixColor = "|cFFFF0000"
                            elseif iLevelDiff == 0 then
                                prefixColor = "|cFFFFFFFF"
                            else
                                prefixColor = "|cFF00FF00"
                            end

                            requestFrameScrollItem.labelItemLvlIncrease:SetFormattedText(prefixColor .. tostring(iLevelDiff))
                        end
                    end


                else
                    requestFrameScrollItem.textureAndLabelWrapper:EnableMouse(false)
                    requestFrameScrollItem.textureItem:SetTexture("Interface\\Icons\\INV_Misc_QuestionMark")
                    requestFrameScrollItem.labelItem:SetText("Item info pending from server...")
                    requestFrameScrollItem.labelItemLvlIncrease:SetText("?")
                end



            else
                requestFrameScrollItem.textureAndLabelWrapper:EnableMouse(false)
                requestFrameScrollItem.textureItem:SetTexture("Interface\\Icons\\INV_Misc_QuestionMark")
                requestFrameScrollItem.labelItem:SetText("No item linked")
                requestFrameScrollItem.labelItemLvlIncrease:SetText("?")
            end



            if isNeed then
                requestFrameScrollItem.labelNeedGreed:SetFormattedText("|cFF00FF00Item was needed")
            else
                requestFrameScrollItem.labelNeedGreed:SetFormattedText("|cFFFFFF00Item was greeded")
            end


            requestFrameScrollItem.awardButton:RegisterForClicks("LeftButtonUp")
            requestFrameScrollItem.awardButton:SetScript("OnClick",
                function(widget)
                    StaticPopupDialogs["LOOTSQUID_AWARD_LOOT"].text = "Really transfer item to: " .. k
                    local popup = StaticPopup_Show("LOOTSQUID_AWARD_LOOT", k)
                    if popup then
                        popup:SetFrameStrata("FULLSCREEN_DIALOG")
                        popup.playerName = k
                    end
                end)

            requestFrameScrollItem.rollButton:RegisterForClicks("LeftButtonUp")
            requestFrameScrollItem.rollButton:SetScript("OnClick",
                function(widget)
                    LootSquid:WhisperToPlayer(k,
                        "The loot master has requested that you please perform a random roll as part of the loot distribution.")
                end)
        end

        requestFrameScrollItemsIndex = requestFrameScrollItemsIndex + 1
    end

    return requestFrameScrollItemsIndex
end

function LootSquid:WhisperToPlayer(playerName, message)
    if LootSquid.debug.disableWhispers then
        return
    end

    SendChatMessage("LootSquid: " .. tostring(message),
        "WHISPER",
        nil,
        playerName)
end

function LootSquid:RaidWarning(message)
    SendChatMessage("LootSquid: " .. tostring(message),
        "RAID_WARNING")
end

function LootSquid:UpdateRequestFrame()

    LootSquid.lastRequestFrameUpdatedTime = GetTime()

    LootSquid:LogDebug("UpdateRequestFrame")

    local offset = self:AddStandingsDataRows(0, self.session.playersNeeding, true)

    if offset < 0 then
        self:LogError("offset was below zero for session.playersNeeding, this is not supposed to happen! Bailing out.")
        return
    end

    offset = self:AddStandingsDataRows(offset, self.session.playersGreeding, false)

    if offset < 1 then
        self.frames.requestFrameScroll.frameMessage:SetText("No requests have been received yet.")
        self.frames.requestFrameScroll.frameMessage:Show()
    else
        self.frames.requestFrameScroll.frameMessage:Hide()
    end

    LootSquid.frames.requestFrameScroll:HideIndexAndAfter(self.frames.requestFrameScroll, offset)
end

function LootSquid:ClearNeedAndGreedRequests()
    self.session.playersNeeding = {}
    self.session.playersGreeding = {}
    LootSquid:UpdateRequestFrame()
    _G.LootSquid.session.currentLootSlot = 0

    if self.frames.overviewCurrentSessionItemName then
        self.frames.overviewCurrentSessionItemName:SetText("No session is active")
    end
end